package net.sf.saxon.instruct;

import net.sf.saxon.expr.*;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.SingletonIterator;
import net.sf.saxon.xpath.XPathException;

import java.util.ArrayList;
import java.util.List;

/**
 * This class is an abstract superclass for Instructions that also behave as Expressions.
 * Currently this is restricted to those instructions used when an XQuery query is compiled.
 * Eventually it is intended that all instructions should support the Expression interface.
 * Note that the Expression interfaces (e.g. iterate()) may only be used on an ExprInstruction
 * if all its child instructions are also instances of ExprInstruction.
 */

public abstract class ExprInstruction extends Instruction implements Expression {

    /**
     * Simplify an expression. This performs any static optimization (by rewriting the expression
     * as a different expression). The default implementation does nothing.
     *
     * @exception XPathException if an error is discovered during expression
     *     rewriting
     * @return the simplified expression
     */

     public Expression simplify() throws XPathException {
        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof Expression) {
                    Expression exp = ((Expression)children[c]).simplify();
                    if (exp instanceof Instr) {
                        children[c] = (Instr)exp;
                    } else {
                        SequenceInstruction seq = new SequenceInstruction(exp, null);
                        children[c] = seq;
                    }
                }
            }
        }
        return this;
    }

    /**
     * Perform static analysis of an expression and its subexpressions.
     *
     * <p>This checks statically that the operands of the expression have
     * the correct type; if necessary it generates code to do run-time type checking or type
     * conversion. A static type error is reported only if execution cannot possibly succeed, that
     * is, if a run-time type error is inevitable. The call may return a modified form of the expression.</p>
     *
     * <p>This method is called after all references to functions and variables have been resolved
     * to the declaration of the function or variable. However, the types of such functions and
     * variables will only be accurately known if they have been explicitly declared.</p>
     *
     * @param env the static context of the expression
     * @exception net.sf.saxon.xpath.XPathException if an error is discovered during this phase
     *     (typically a type error)
     * @return the original expression, rewritten to perform necessary
     *     run-time type checks, and to perform other type-related
     *     optimizations
     */

    public Expression analyze(StaticContext env) throws XPathException {
        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof Expression) {
                    Expression exp = ((Expression)children[c]).analyze(env);
                    if (exp instanceof Instr) {
                        children[c] = (Instr)exp;
                    } else {
                        SequenceInstruction seq = new SequenceInstruction(exp, null);
                        children[c] = seq;
                    }
                }
            }
        }
        return this;
    }

    /**
     * Determine which aspects of the context the expression depends on. The result is
     * a bitwise-or'ed value composed from constants such as {@link StaticProperty#DEPENDS_ON_CONTEXT_ITEM} and
     * {@link StaticProperty#DEPENDS_ON_CURRENT_ITEM}. The default implementation combines the intrinsic
     * dependencies of this expression with the dependencies of the subexpressions,
     * computed recursively. This is overridden for expressions such as FilterExpression
     * where a subexpression's dependencies are not necessarily inherited by the parent
     * expression.
     *
     * @return a set of bit-significant flags identifying the dependencies of
     *     the expression
     */

    public int getDependencies() {
        int dep = 0;
        Expression[] sub = getSubExpressions();
        for (int i=0; i<sub.length; i++) {
            dep |= sub[i].getDependencies();
        }
        return dep;
    }

    /**
     * Get the static properties of this expression (other than its type). The result is
     * bit-signficant. These properties are used for optimizations. In general, if
     * property bit is set, it is true, but if it is unset, the value is unknown.
     *
     * @return a set of flags indicating static properties of this expression
     */

    public int getSpecialProperties() {
        return StaticProperty.CREATES_NEW_NODES;
        // TODO: we can do better than this!
    }

    /**
     * Handle promotion offers, that is, non-local tree rewrites.
     * @param offer The type of rewrite being offered
     * @throws XPathException
     */

    protected abstract void promoteInst(PromotionOffer offer) throws XPathException;

    /**
     * Offer promotion for this subexpression. The offer will be accepted if the subexpression
     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
     * This method is always called at compile time.
     *
     * @param offer details of the offer, for example the offer to move
     *     expressions that don't depend on the context to an outer level in
     *     the containing expression
     * @exception XPathException if any error is detected
     * @return if the offer is not accepted, return this expression unchanged.
     *      Otherwise return the result of rewriting the expression to promote
     *      this subexpression
     */

    public Expression promote(PromotionOffer offer) throws XPathException {
        Expression exp = offer.accept(this);
        if (exp!=null) {
            return exp;
        } else {
            promoteInst(offer);
            return this;
        }
    }

    /**
     * Get all the XPath expressions associated with this instruction
     * (in XSLT terms, the expression present on attributes of the instruction,
     * as distinct from the child instructions in a sequence construction)
     * @param list A list to be populated with the list of XPath expressions
     */

    protected abstract void getXPathExpressions(List list);

    /**
      * Get the immediate sub-expressions of this expression.
      * @return an array containing the sub-expressions of this expression
      */

    public Expression[] getSubExpressions() {
        List list = new ArrayList();
        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof Expression) {
                    list.add(children[c]);
                } else {
                    throw new IllegalStateException("Children of an ExprInstruction must themselves be Expressions");
                }
            }
        }
        getXPathExpressions(list);
        Expression[] array = new Expression[list.size()];
        for (int i=0; i<list.size(); i++) {
            array[i] = (Expression)list.get(i);
        }
        return array;
    }

    /**
     * Evaluate an expression as a single item. This always returns either a single Item or
     * null (denoting the empty sequence). No conversion is done. This method should not be
     * used unless the static type of the expression is a subtype of "item" or "item?": that is,
     * it should not be called if the expression may return a sequence. There is no guarantee that
     * this condition will be detected.
     *
     * @param context The context in which the expression is to be evaluated
     * @exception XPathException if any dynamic error occurs evaluating the
     *     expression
     * @return the node or atomic value that results from evaluating the
     *     expression; or null to indicate that the result is an empty
     *     sequence
     */

    public Item evaluateItem(XPathContext context) throws XPathException {
        return iterate(context).next();
    }

   /**
     * Evaluate an expression as a String. This function must only be called in contexts
     * where it is known that the expression will return a single string (or where an empty sequence
     * is to be treated as a zero-length string). Implementations should not attempt to convert
     * the result to a string, other than converting () to "". This method is used mainly to
     * evaluate expressions produced by compiling an attribute value template.
     *
     * @exception XPathException if any dynamic error occurs evaluating the
     *     expression
     * @exception ClassCastException if the result type of the
     *     expression is not xs:string?
     * @param context The context in which the expression is to be evaluated
     * @return the value of the expression, evaluated in the current context.
     *     The expression must return a string or (); if the value of the
     *     expression is (), this method returns "".
     */

    public final String evaluateAsString(XPathContext context) throws XPathException {
        Item item = evaluateItem(context);
        if (item==null) {
            return "";
        } else {
            return item.getStringValue();
        }
    }

    /**
     * Get the effective boolean value of the expression. This returns false if the value
     * is the empty sequence, a zero-length string, a number equal to zero, or the boolean
     * false. Otherwise it returns true.
     *
     * @param context The context in which the expression is to be evaluated
     * @exception net.sf.saxon.xpath.XPathException if any dynamic error occurs evaluating the
     *     expression
     * @return the effective boolean value
     */


    public boolean effectiveBooleanValue(XPathContext context) throws XPathException {
        return ExpressionTool.effectiveBooleanValue(iterate(context));
    }

    /**
     * Return an Iterator to iterate over the values of a sequence. The value of every
     * expression can be regarded as a sequence, so this method is supported for all
     * expressions. This default implementation handles iteration for expressions that
     * return singleton values: for non-singleton expressions, the subclass must
     * provide its own implementation.
     *
     * @exception net.sf.saxon.xpath.XPathException if any dynamic error occurs evaluating the
     *     expression
     * @param context supplies the context for evaluation
     * @return a SequenceIterator that can be used to iterate over the result
     *     of the expression
     */

    public SequenceIterator iterate(XPathContext context) throws XPathException {
        Item item = evaluateItem(context);
        return SingletonIterator.makeIterator(item);
    }

}
