package net.sf.saxon.instruct;

import net.sf.saxon.Controller;
import net.sf.saxon.type.Type;
import net.sf.saxon.trace.InstructionInfo;
import net.sf.saxon.trace.TraceListener;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.UserFunctionCall;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.value.Value;
import net.sf.saxon.value.ObjectValue;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerException;
import javax.xml.transform.SourceLocator;
import java.util.HashMap;
import java.io.Serializable;

/**
 * This object represents the compiled form of a user-written function
 * (the source can be either an XSLT stylesheet function or an XQuery function).
 *
 * <p>It is assumed that type-checking, of both the arguments and the results,
 * has been handled at compile time. That is, the expression supplied as the body
 * of the function must be wrapped in code to check or convert the result to the
 * required type, and calls on the function must be wrapped at compile time to check or
 * convert the supplied arguments.
 */

public final class UserFunction implements CallableFunction,
        Serializable, InstructionInfo, SourceLocator {

    private Expression body;
    private String functionName;
    private String systemId;
    private boolean memoFunction = false;
    private int lineNumber;


    public UserFunction(Expression body) {
        this.body = body;
    };

    public void setLineNumber(int lineNumber) {
        this.lineNumber = lineNumber;
    }

    public void setSystemId(String systemId) {
        this.systemId = systemId;
    }

    public int getLineNumber() {
        return lineNumber;
    }

    public String getSystemId() {
        return systemId;
    }

    public int getColumnNumber() {
        return -1;
    }

    public String getPublicId() {
        return null;
    }

    public String getInstructionName() {
        return "function";
    }

    public Object getProperty(String propertyName) {
        if (propertyName.equals("name")) {
            return functionName;
        } else {
            return null;
        }
    }

    public HashMap getProperties() {
        HashMap map = new HashMap();
        map.put("name", functionName);
        return map;
    }

    public void setMemoFunction(boolean isMemo) {
        // TODO: the memo functionality of this class is never invoked
        memoFunction = isMemo;
    }

    public void setFunctionName(String name) {
        functionName = name;
    }

    public String getFunctionName() {
        return functionName;
    }

    /**
     * Call this function.
     * @param actualArgs the arguments supplied to the function. These must have the correct
     * types required by the function signature (it is the caller's responsibility to check this).
     * It is acceptable to supply a {@link net.sf.saxon.value.Closure} to represent a value whose
     * evaluation will be delayed until it is needed. The array must be the correct size to match
     * the number of arguments: again, it is the caller's responsibility to check this.
     * @param controller This provides the run-time context for evaluating the function.
     * @param evaluateTailCalls if true, then any function calls contained in the body of the function
     * are evaluated in the normal way, whether or not they are marked as tail calls. If the argument
     * is false, then tail calls are not evaluated, and instead a FunctionCallPackage is returned containing
     * the information needed to evaluate the function. The caller must then be prepared to deal with this
     * returned value by evaluating the packaged function call (which may return further packaged function
     * calls, and so on).
     * @return a Value representing the result of the function.
     */

    public Value call(Value[] actualArgs, Controller controller, boolean evaluateTailCalls)
            throws TransformerException {

        // If this is a memo function, see if the result is already known
        if (memoFunction) {
            Value value = getCachedValue(controller, actualArgs);
            if (value != null) return value;
        }

        // Otherwise evaluate the function

        if (controller.isTracing()) {
            TraceListener listener = controller.getTraceListener();
            listener.enter(this);
        }

        Bindery bindery = controller.getBindery();
        bindery.openStackFrame(actualArgs);
        Value result;
        try {
            result = ExpressionTool.eagerEvaluate(body, controller.newXPathContext());
            //TODO: lazy evaluation currently doesn't work here. (E.g. tour.xq). Find out why not.
        } catch (XPathException err) {
            err.setLocator(this);
            throw err;
        }

        if (evaluateTailCalls) {
            while (result instanceof ObjectValue &&
                    ((ObjectValue)result).getObject() instanceof UserFunctionCall.FunctionCallPackage) {
                result = ((UserFunctionCall.FunctionCallPackage)((ObjectValue)result).getObject()).call();
            }
        }
        bindery.closeStackFrame();

        if (controller.isTracing()) {
            TraceListener listener = controller.getTraceListener();
            listener.leave(this);
        }

        // If this is a memo function, save the result in the cache
        if (memoFunction) {
            putCachedValue(controller, actualArgs, result);
        }

        return result;
    }

    /**
     * Call this function. This method allows an XQuery function to be called directly from a Java
     * application. It creates the environment needed to achieve this
     * @param actualArgs the arguments supplied to the function. These must have the correct
     * types required by the function signature (it is the caller's responsibility to check this).
     * It is acceptable to supply a {@link net.sf.saxon.value.Closure} to represent a value whose
     * evaluation will be delayed until it is needed. The array must be the correct size to match
     * the number of arguments: again, it is the caller's responsibility to check this.
     * @param controller This provides the run-time context for evaluating the function. A Controller
     * may be obtained by calling {@link net.sf.saxon.query.XQueryExpression#getController}. This may
     * be used for a series of calls on functions defined in the same module as the XQueryExpression.
     * @return a Value representing the result of the function.
     */

    public Value call(Value[] actualArgs, Controller controller) throws TransformerException {
        return call(actualArgs, controller, true);
    }

    /**
     * For memo functions, get a saved value from the cache.
     * @return the cached value, or null if no value has been saved for these parameters
     */

    private Value getCachedValue(Controller controller, Value[] params) {

        try {
            HashMap map = (HashMap) controller.getUserData(this, "memo-function-cache");
            if (map == null) {
                return null;
            }
            String key = getCombinedKey(params);
            return (Value) map.get(key);
        } catch (XPathException err) {
            return null;
        }
    }

    /**
     * For memo functions, put the computed value in the cache.
     */

    private void putCachedValue(Controller controller, Value[] params, Value value) {
        try {
            HashMap map = (HashMap) controller.getUserData(this, "memo-function-cache");
            if (map == null) {
                map = new HashMap();
                controller.setUserData(this, "memo-function-cache", map);
            }
            String key = getCombinedKey(params);
            map.put(key, value);
        } catch (XPathException err) {
            // it doesn't matter if we fail to cache the result
        }
    }

    /**
     * Get a key value representing the values of all the supplied arguments
     */

    private static String getCombinedKey(Value[] params) throws XPathException {
        StringBuffer sb = new StringBuffer();

        for (int i = 0; i < params.length; i++) {
            Value val = params[i];
            SequenceIterator iter = val.iterate(null);
            while (true) {
                Item item = iter.next();
                if (item == null) {
                    break;
                }
                if (item instanceof NodeInfo) {
                    NodeInfo node = (NodeInfo) item;
                    //sb.append(""+node.getDocumentRoot().getDocumentNumber());
                    //sb.append('/');
                    sb.append(node.generateId());
                } else {
                    sb.append("" + Type.displayTypeName(item));
                    sb.append('/');
                    sb.append(item.getStringValue());
                }
                sb.append('\u0001');
            }
            sb.append('\u0002');
        }
        return sb.toString();
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Michael H. Kay
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s):
// Portions marked "e.g." are from Edwin Glaser (edwin@pannenleiter.de)
//

