package net.sf.saxon.sort;
import net.sf.saxon.value.*;
import net.sf.saxon.type.Type;

import java.util.Comparator;
import java.text.Collator;
import java.text.CollationKey;

/**
 * A Comparator used for comparing atomic values of arbitrary item types. It encapsulates
 * a Collator that is used when the values to be compared are strings. It also supports
 * a separate method for testing equality of items, which can be used for data types that
 * are not ordered.
 *
 * The AtomicSortComparer is identical to the AtomicComparer except for its handling
 * of NaN: it treats NaN values as lower than any other value, and NaNs as equal to
 * each other.
 *
 * @author Michael H. Kay
 *
 */

public class AtomicSortComparer implements Comparator, java.io.Serializable {

    private Comparator collator;

    public AtomicSortComparer(Comparator collator) {
        this.collator = collator;
        if (collator == null) {
            this.collator = CodepointCollator.getInstance();
        }
    }

    /**
    * Compare two AtomicValue objects according to the rules for their data type. UntypedAtomic
    * values are compared as if they were strings; if different semantics are wanted, the conversion
    * must be done by the caller.
    * @param a the first object to be compared. It is intended that this should be an instance
    * of AtomicValue, though this restriction is not enforced. If it is a StringValue, the
    * collator is used to compare the values, otherwise the value must implement the java.util.Comparable
    * interface.
    * @param b the second object to be compared. This must be comparable with the first object: for
    * example, if one is a string, they must both be strings.
    * @return <0 if a<b, 0 if a=b, >0 if a>b
    * @throws ClassCastException if the objects are not comparable
    */

    public int compare(Object a, Object b) {

        // System.err.println("Comparing " + a.getClass() + "(" + a + ") with " + b.getClass() + "(" + b + ") using " + collator);
        if (a instanceof DerivedAtomicValue) {
            a = ((DerivedAtomicValue)a).getPrimitiveValue();
        }
        if (b instanceof DerivedAtomicValue) {
            b = ((DerivedAtomicValue)b).getPrimitiveValue();
        }
        if (a instanceof UntypedAtomicValue) {
            return ((UntypedAtomicValue)a).compareTo(b, collator);
        } else if (b instanceof UntypedAtomicValue) {
            return -((UntypedAtomicValue)b).compareTo(a, collator);
        } else if (a instanceof DoubleValue && Double.isNaN(((DoubleValue)a).asDouble())) {
            if (b instanceof DoubleValue && Double.isNaN(((DoubleValue)b).asDouble())) {
                return 0;
            } else {
                return -1;
            }
        } else if (b instanceof DoubleValue && Double.isNaN(((DoubleValue)b).asDouble())) {
            return +1;
        } else if (a instanceof Comparable) {
            return ((Comparable)a).compareTo(b);
        } else if (a instanceof StringValue) {
            return collator.compare(((StringValue)a).getStringValue(), ((StringValue)b).getStringValue());
        } else {
            throw new ClassCastException("Objects are not comparable (" + a.getClass() + ", " + b.getClass() + ")");
        }
    }

    /**
    * Compare two AtomicValue objects for equality according to the rules for their data type. UntypedAtomic
    * values are compared as if they were strings; if different semantics are wanted, the conversion
    * must be done by the caller.
    * @param a the first object to be compared. It is intended that this should be an instance
    * of AtomicValue, though this restriction is not enforced. If it is a StringValue, the
    * collator is used to compare the values, otherwise the value must implement the equals() method.
    * @param b the second object to be compared. This must be comparable with the first object: for
    * example, if one is a string, they must both be strings.
    * @return <0 if a<b, 0 if a=b, >0 if a>b
    * @throws ClassCastException if the objects are not comparable
    */

    public boolean comparesEqual(Object a, Object b) {
        // System.err.println("Comparing " + a.getClass() + ": " + a + " with " + b.getClass() + ": " + b);
        if (a instanceof UntypedAtomicValue) {
            return ((UntypedAtomicValue)a).compareTo(b, collator) == 0;
        } else if (b instanceof UntypedAtomicValue) {
            return ((UntypedAtomicValue)b).compareTo(a, collator) == 0;
        } else if (a instanceof StringValue) {
            return collator.compare(((StringValue)a).getStringValue(), ((StringValue)b).getStringValue()) == 0;
        } else if (a instanceof String) {
            return collator.compare(a, b) == 0;
        } else if (a instanceof DoubleValue && Double.isNaN(((DoubleValue)a).asDouble())) {
            return (b instanceof DoubleValue && Double.isNaN(((DoubleValue)a).asDouble()));
        } else {
            return a.equals(b);
        }
    }

    /**
    * Get a comparison key for an object. This must satisfy the rule that if two objects are equal,
    * then their comparison keys are equal, and vice versa. There is no requirement that the
    * comparison keys should reflect the ordering of the underlying objects.
    */

    public ComparisonKey getComparisonKey(AtomicValue a) {
        AtomicValue prim = a.getPrimitiveValue();
        if (prim instanceof NumericValue) {
            if (((NumericValue)prim).isNaN()) {
                // Deal with NaN specially. For this function, NaN is considered equal to itself
                return new ComparisonKey(Type.NUMBER, new StringValue("NaN"));
            } else {
                return new ComparisonKey(Type.NUMBER, prim);
            }
        } else if (prim instanceof StringValue) {
            if (collator instanceof Collator) {
                return new ComparisonKey(Type.STRING,
                        ((Collator)collator).getCollationKey(((StringValue)prim).getStringValue()));
            } else {
                return new ComparisonKey(Type.STRING, prim);
            }
        } else {
            return new ComparisonKey(prim.getItemType().getPrimitiveType(), prim);
        }
    }

    public static class ComparisonKey {
        int category;
        Object value;

        public ComparisonKey(int category, AtomicValue value) {
            this.category = category;
            this.value = value;
        }

        public ComparisonKey(int category, CollationKey value) {
            this.category = category;
            this.value = value;
        }

        public boolean equals(Object other) {
            if (other instanceof ComparisonKey) {
                ComparisonKey otherKey = (ComparisonKey)other;
                return this.category == otherKey.category &&
                        this.value.equals(otherKey.value);
            } else {
                throw new ClassCastException("Cannot compare a ComparisonKey to an object of a different class");
            }
        }

        public int hashCode() {
            return value.hashCode() ^ category;
        }

    }

}
